/****************************************************************************
 * Visual Boy Advance GX
 *
 * Tantric 2008-2009
 *
 * fileop.cpp
 *
 * File operations
 ***************************************************************************/

#include <gccore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ogcsys.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <zlib.h>
#include <malloc.h>
#include <sdcard/wiisd_io.h>
#include <sdcard/gcsd.h>
#include <ogc/usbstorage.h>
#include <ogc/dvd.h>

#include "vbagx.h"
#include "vbasupport.h"
#include "fileop.h"
#include "networkop.h"
#include "gcunzip.h"
#include "menu.h"
#include "filebrowser.h"
#include "gui/gui.h"

#define THREAD_SLEEP 100

unsigned char *savebuffer;
static mutex_t bufferLock = LWP_MUTEX_NULL;
FILE * file; // file pointer - the only one we should ever use!
bool unmountRequired[7] = { false, false, false, false, false, false, false };
bool isMounted[7] = { false, false, false, false, false, false, false };

#ifdef HW_RVL
	const DISC_INTERFACE* sd = &__io_wiisd;
	const DISC_INTERFACE* usb = &__io_usbstorage;
//	const DISC_INTERFACE* dvd = &__io_wiidvd;
#else
	const DISC_INTERFACE* carda = &__io_gcsda;
	const DISC_INTERFACE* cardb = &__io_gcsdb;
	const DISC_INTERFACE* dvd = &__io_gcdvd;
#endif

// folder parsing thread
static lwp_t parsethread = LWP_THREAD_NULL;
static DIR_ITER * dirIter = NULL;
static bool parseHalt = true;
static bool parseFilter = true;
bool ParseDirEntries();
int selectLoadedFile = 0;

// device thread
static lwp_t devicethread = LWP_THREAD_NULL;
static bool deviceHalt = true;

/****************************************************************************
 * ResumeDeviceThread
 *
 * Signals the device thread to start, and resumes the thread.
 ***************************************************************************/
void
ResumeDeviceThread()
{
	deviceHalt = false;
	LWP_ResumeThread(devicethread);
}

/****************************************************************************
 * HaltGui
 *
 * Signals the device thread to stop.
 ***************************************************************************/
void
HaltDeviceThread()
{
#ifdef HW_RVL
	deviceHalt = true;

	// wait for thread to finish
	while(!LWP_ThreadIsSuspended(devicethread))
		usleep(THREAD_SLEEP);
#endif
}

/****************************************************************************
 * HaltParseThread
 *
 * Signals the parse thread to stop.
 ***************************************************************************/
void
HaltParseThread()
{
	parseHalt = true;

	while(!LWP_ThreadIsSuspended(parsethread))
		usleep(THREAD_SLEEP);
}


/****************************************************************************
 * devicecallback
 *
 * This checks our devices for changes (SD/USB/DVD removed)
 ***************************************************************************/
#ifdef HW_RVL
static int devsleep;

static void *
devicecallback (void *arg)
{
	while (1)
	{
		if(isMounted[DEVICE_SD])
		{
			if(!sd->isInserted()) // check if the device was removed
			{
				unmountRequired[DEVICE_SD] = true;
				isMounted[DEVICE_SD] = false;
			}
		}

		if(isMounted[DEVICE_USB])
		{
			if(!usb->isInserted()) // check if the device was removed
			{
				unmountRequired[DEVICE_USB] = true;
				isMounted[DEVICE_USB] = false;
			}
		}

		/*if(isMounted[DEVICE_DVD])
		{
			if(!dvd->isInserted()) // check if the device was removed
			{
				unmountRequired[DEVICE_DVD] = true;
				isMounted[DEVICE_DVD] = false;
			}
		}*/

		devsleep = 1000*1000; // 1 sec

		while(devsleep > 0)
		{
			if(deviceHalt)
				LWP_SuspendThread(devicethread);
			usleep(THREAD_SLEEP);
			devsleep -= THREAD_SLEEP;
		}
		//UpdateCheck();
	}
	return NULL;
}
#endif

static void *
parsecallback (void *arg)
{
	while(1)
	{
		while(ParseDirEntries())
			usleep(THREAD_SLEEP);
		LWP_SuspendThread(parsethread);
	}
	return NULL;
}

/****************************************************************************
 * InitDeviceThread
 *
 * libOGC provides a nice wrapper for LWP access.
 * This function sets up a new local queue and attaches the thread to it.
 ***************************************************************************/
void
InitDeviceThread()
{
#ifdef HW_RVL
	LWP_CreateThread (&devicethread, devicecallback, NULL, NULL, 0, 40);
#endif
	LWP_CreateThread (&parsethread, parsecallback, NULL, NULL, 0, 80);
}

/****************************************************************************
 * UnmountAllFAT
 * Unmounts all FAT devices
 ***************************************************************************/
void UnmountAllFAT()
{
#ifdef HW_RVL
	fatUnmount("sd:");
	fatUnmount("usb:");
#else
	fatUnmount("carda:");
	fatUnmount("cardb:");
#endif
}

/****************************************************************************
 * MountFAT
 * Checks if the device needs to be (re)mounted
 * If so, unmounts the device
 * Attempts to mount the device specified
 * Sets libfat to use the device by default
 ***************************************************************************/

static bool MountFAT(int device, int silent)
{
	bool mounted = false;
	int retry = 1;
	char name[10], name2[10];
	const DISC_INTERFACE* disc = NULL;

	switch(device)
	{
#ifdef HW_RVL
		case DEVICE_SD:
			sprintf(name, "sd");
			sprintf(name2, "sd:");
			disc = sd;
			break;
		case DEVICE_USB:
			sprintf(name, "usb");
			sprintf(name2, "usb:");
			disc = usb;
			break;
#else
		case DEVICE_SD_SLOTA:
			sprintf(name, "carda");
			sprintf(name2, "carda:");
			disc = carda;
			break;

		case DEVICE_SD_SLOTB:
			sprintf(name, "cardb");
			sprintf(name2, "cardb:");
			disc = cardb;
			break;
#endif
		default:
			return false; // unknown device
	}

	if(unmountRequired[device])
	{
		unmountRequired[device] = false;
		fatUnmount(name2);
		disc->shutdown();
		isMounted[device] = false;
	}

	while(retry)
	{
		if(disc->startup() && fatMountSimple(name, disc))
			mounted = true;

		if(mounted || silent)
			break;

#ifdef HW_RVL
		if(device == DEVICE_SD)
			retry = ErrorPromptRetry("SD card not found!");
		else
			retry = ErrorPromptRetry("USB drive not found!");
#else
		retry = ErrorPromptRetry("SD card not found!");
#endif
	}

	isMounted[device] = mounted;
	return mounted;
}

void MountAllFAT()
{
#ifdef HW_RVL
	MountFAT(DEVICE_SD, SILENT);
	MountFAT(DEVICE_USB, SILENT);
#else
	MountFAT(DEVICE_SD_SLOTA, SILENT);
	MountFAT(DEVICE_SD_SLOTB, SILENT);
#endif
}

/****************************************************************************
 * MountDVD()
 *
 * Tests if a ISO9660 DVD is inserted and available, and mounts it
 ***************************************************************************/
bool MountDVD(bool silent)
{
	bool mounted = false;
	int retry = 1;

	//dont support multiple insert/removal of DVD
	if( FSTIsMounted() )
	    return true;

	while(retry)
	{
		//ShowAction("Loading DVD...");

		if(!FST_Mount())
		{
			if(silent)
				break;

			retry = ErrorPromptRetry("FST Mount Error!");
		}
		else
		{
			mounted = true;
			break;
		}
	}
	//CancelAction();
	isMounted[DEVICE_DVD] = mounted;
	return mounted;
}

bool FindDevice(char * filepath, int * device)
{
	if(!filepath || filepath[0] == 0)
		return false;

	if(strncmp(filepath, "sd:", 3) == 0)
	{
		*device = DEVICE_SD;
		return true;
	}
	else if(strncmp(filepath, "usb:", 4) == 0)
	{
		*device = DEVICE_USB;
		return true;
	}
	else if(strncmp(filepath, "smb:", 4) == 0)
	{
		*device = DEVICE_SMB;
		return true;
	}
	else if(strncmp(filepath, "carda:", 6) == 0)
	{
		*device = DEVICE_SD_SLOTA;
		return true;
	}
	else if(strncmp(filepath, "cardb:", 6) == 0)
	{
		*device = DEVICE_SD_SLOTB;
		return true;
	}
	else if(strncmp(filepath, "dvd:", 4) == 0)
	{
		*device = DEVICE_DVD;
		return true;
	}
	return false;
}

char * StripDevice(char * path)
{
	if(path == NULL)
		return NULL;
	
	char * newpath = strchr(path,'/');
	
	if(newpath != NULL)
		newpath++;
	
	return newpath;
}

/****************************************************************************
 * ChangeInterface
 * Attempts to mount/configure the device specified
 ***************************************************************************/
bool ChangeInterface(int device, bool silent)
{
	if(isMounted[device])
		return true;

	bool mounted = false;

	switch(device)
	{
#ifdef HW_RVL
		case DEVICE_SD:
		case DEVICE_USB:
#else
		case DEVICE_SD_SLOTA:
		case DEVICE_SD_SLOTB:
#endif
			mounted = MountFAT(device, silent);
			break;
		case DEVICE_DVD:
			mounted = MountDVD(silent);
			break;
#ifdef HW_RVL
		case DEVICE_SMB:
			mounted = ConnectShare(silent);
			break;
#endif
	}

	return mounted;
}

bool ChangeInterface(char * filepath, bool silent)
{
	int device = -1;

	if(!FindDevice(filepath, &device))
		return false;

	return ChangeInterface(device, silent);
}

void CreateAppPath(char * origpath)
{
	if(!origpath || origpath[0] == 0)
		return;

	char * path = strdup(origpath); // make a copy so we don't mess up original

	if(!path)
		return;
	
	char * loc = strrchr(path,'/');
	if (loc != NULL)
		*loc = 0; // strip file name

	int pos = 0;

	// replace fat:/ with sd:/
	if(strncmp(path, "fat:/", 5) == 0)
	{
		pos++;
		path[1] = 's';
		path[2] = 'd';
	}
	if(ChangeInterface(&path[pos], SILENT))
		strncpy(appPath, &path[pos], MAXPATHLEN);
	appPath[MAXPATHLEN-1] = 0;
	free(path);
}

static char *GetExt(char *file)
{
	if(!file)
		return NULL;

	char *ext = strrchr(file,'.');
	if(ext != NULL)
	{
		ext++;
		int extlen = strlen(ext);
		if(extlen > 5)
			return NULL;
	}
	return ext;
}

bool ParseDirEntries()
{
	if(!dirIter)
		return false;

	char filename[MAXPATHLEN];
	char *ext;
	struct stat filestat;

	int i = 0;
	int res;

	while(i < 20)
	{
		res = dirnext(dirIter,filename,&filestat);

		if(res != 0)
			break;

		if(filename[0] == '.' && filename[1] != '.')
			continue;

		ext = GetExt(filename);

		// don't show the file if it's not a valid ROM
		if(parseFilter && (filestat.st_mode & _IFDIR) == 0)
		{
			if(ext == NULL)
				continue;

			if(	stricmp(ext, "agb") != 0 && stricmp(ext, "gba") != 0 &&
				stricmp(ext, "bin") != 0 && stricmp(ext, "elf") != 0 &&
				stricmp(ext, "mb") != 0 && stricmp(ext, "dmg") != 0 &&
				stricmp(ext, "gb") != 0 && stricmp(ext, "gbc") != 0 &&
				stricmp(ext, "cgb") != 0 && stricmp(ext, "sgb") != 0 &&
				stricmp(ext, "zip") != 0 && stricmp(ext, "7z") != 0)
				continue;
		}

		if(!AddBrowserEntry())
		{
			parseHalt = true;
			break;
		}

		strncpy(browserList[browser.numEntries+i].filename, filename, MAXJOLIET);
		browserList[browser.numEntries+i].length = filestat.st_size;
		browserList[browser.numEntries+i].mtime = filestat.st_mtime;
		browserList[browser.numEntries+i].isdir = (filestat.st_mode & _IFDIR) == 0 ? 0 : 1; // flag this as a dir

		if(browserList[browser.numEntries+i].isdir)
		{
			if(strcmp(filename, "..") == 0)
				sprintf(browserList[browser.numEntries+i].displayname, "Up One Level");
			else
				strncpy(browserList[browser.numEntries+i].displayname, browserList[browser.numEntries+i].filename, MAXJOLIET);
			browserList[browser.numEntries+i].icon = ICON_FOLDER;
		}
		else
		{
			StripExt(browserList[browser.numEntries+i].displayname, browserList[browser.numEntries+i].filename); // hide file extension
		}
		i++;
	}

	// Sort the file list
	if(i >= 0)
	{
		qsort(browserList, browser.numEntries+i, sizeof(BROWSERENTRY), FileSortCallback);
	}

	// try to find and select the last loaded file
	if(selectLoadedFile == 1 && res != 0 && loadedFile[0] != 0 && browser.dir[0] != 0)
	{
		int indexFound = -1;
		
		for(int j=1; j < browser.numEntries + i; j++)
		{
			if(strcmp(browserList[j].filename, loadedFile) == 0)
			{
				indexFound = j;
				break;
			}
		}

		// move to this file
		if(indexFound > 0)
		{
			if(indexFound > FILE_PAGESIZE)
			{			
				browser.pageIndex = (ceil(indexFound/FILE_PAGESIZE*1.0)) * FILE_PAGESIZE;
				
				if(browser.pageIndex + FILE_PAGESIZE > browser.numEntries + i)
				{
					browser.pageIndex = browser.numEntries + i - FILE_PAGESIZE;
				}
			}
			browser.selIndex = indexFound;
		}
		selectLoadedFile = 2; // selecting done
	}

	browser.numEntries += i;

	if(res != 0 || parseHalt)
	{
		dirclose(dirIter); // close directory
		dirIter = NULL;
		return false; // no more entries
	}
	return true; // more entries
}

/***************************************************************************
 * Browse subdirectories
 **************************************************************************/
int
ParseDirectory(bool waitParse, bool filter)
{
	int retry = 1;
	bool mounted = false;
	parseFilter = filter;
	
	ResetBrowser(); // reset browser

	// open the directory
	while(dirIter == NULL && retry == 1)
	{
		mounted = ChangeInterface(browser.dir, NOTSILENT);

		if(mounted)
			dirIter = diropen(browser.dir);
		else
			return -1;

		if(dirIter == NULL)
			retry = ErrorPromptRetry("Error opening directory!");
	}

	// if we can't open the dir, try higher levels
	if (dirIter == NULL)
	{
		char * devEnd = strrchr(browser.dir, '/');

		while(!IsDeviceRoot(browser.dir))
		{
			devEnd[0] = 0; // strip slash
			devEnd = strrchr(browser.dir, '/');

			if(devEnd == NULL)
				break;

			devEnd[1] = 0; // strip remaining file listing
			dirIter = diropen(browser.dir);
			if (dirIter)
				break;
		}
	}
	
	if(dirIter == NULL)
		return -1;

	if(IsDeviceRoot(browser.dir))
	{
		AddBrowserEntry();
		sprintf(browserList[0].filename, "..");
		sprintf(browserList[0].displayname, "Up One Level");
		browserList[0].length = 0;
		browserList[0].mtime = 0;
		browserList[0].isdir = 1; // flag this as a dir
		browserList[0].icon = ICON_FOLDER;
		browser.numEntries++;
	}

	parseHalt = false;
	ParseDirEntries(); // index first 20 entries

	LWP_ResumeThread(parsethread); // index remaining entries

	if(waitParse) // wait for complete parsing
	{
		ShowAction("Loading...");

		while(!LWP_ThreadIsSuspended(parsethread))
			usleep(THREAD_SLEEP);

		CancelAction();
	}

	return browser.numEntries;
}

/****************************************************************************
 * AllocSaveBuffer ()
 * Clear and allocate the savebuffer
 ***************************************************************************/
void
AllocSaveBuffer ()
{
	if(bufferLock == LWP_MUTEX_NULL)
		LWP_MutexInit(&bufferLock, false);

	if(bufferLock != LWP_MUTEX_NULL)
		LWP_MutexLock(bufferLock);
	memset (savebuffer, 0, SAVEBUFFERSIZE);
}

/****************************************************************************
 * FreeSaveBuffer ()
 * Free the savebuffer memory
 ***************************************************************************/
void
FreeSaveBuffer ()
{
	if(bufferLock != LWP_MUTEX_NULL)
		LWP_MutexUnlock(bufferLock);
}

/****************************************************************************
 * LoadSzFile
 * Loads the selected file # from the specified 7z into rbuffer
 * Returns file size
 ***************************************************************************/
size_t
LoadSzFile(char * filepath, unsigned char * rbuffer)
{
	size_t size = 0;

	// stop checking if devices were removed/inserted
	// since we're loading a file
	HaltDeviceThread();

	// halt parsing
	HaltParseThread();

	file = fopen (filepath, "rb");
	if (file > 0)
	{
		size = SzExtractFile(browserList[browser.selIndex].filenum, rbuffer);
		fclose (file);
	}
	else
	{
		ErrorPrompt("Error opening file!");
	}

	// go back to checking if devices were inserted/removed
	ResumeDeviceThread();

	return size;
}

/****************************************************************************
 * LoadFile
 ***************************************************************************/
size_t
LoadFile (char * rbuffer, char *filepath, size_t length, bool silent)
{
	char zipbuffer[2048];
	size_t size = 0, offset = 0, readsize = 0;
	int retry = 1;
	int device;

	if(!FindDevice(filepath, &device))
		return 0;

	// stop checking if devices were removed/inserted
	// since we're loading a file
	HaltDeviceThread();

	// halt parsing
	HaltParseThread();

	// open the file
	while(retry)
	{
		if(!ChangeInterface(device, silent))
			break;

		file = fopen (filepath, "rb");

		if(!file)
		{
			if(silent)
				break;

			retry = ErrorPromptRetry("Error opening file!");
			continue;
		}

		if(length > 0 && length <= 2048) // do a partial read (eg: to check file header)
		{
			size = fread (rbuffer, 1, length, file);
		}
		else // load whole file
		{
			readsize = fread (zipbuffer, 1, 32, file);

			if(!readsize)
			{
				unmountRequired[device] = true;
				retry = ErrorPromptRetry("Error reading file!");
				fclose (file);
				continue;
			}

			if (IsZipFile (zipbuffer))
			{
				size = UnZipBuffer ((unsigned char *)rbuffer); // unzip
			}
			else
			{
				fseeko(file,0,SEEK_END);
				size = ftello(file);
				fseeko(file,0,SEEK_SET);

				while(!feof(file))
				{
					ShowProgress ("Loading...", offset, size);
					readsize = fread (rbuffer + offset, 1, 4096, file); // read in next chunk

					if(readsize <= 0)
						break; // reading finished (or failed)

					offset += readsize;
				}
				size = offset;
				CancelAction();
			}
		}
		retry = 0;
		fclose (file);
	}

	// go back to checking if devices were inserted/removed
	ResumeDeviceThread();
	CancelAction();
	return size;
}

size_t LoadFile(char * filepath, bool silent)
{
	return LoadFile((char *)savebuffer, filepath, 0, silent);
}

/****************************************************************************
 * SaveFile
 * Write buffer to file
 ***************************************************************************/
size_t
SaveFile (char * buffer, char *filepath, size_t datasize, bool silent)
{
	size_t written = 0;
	size_t writesize, nextwrite;
	int retry = 1;
	int device;
		
	if(!FindDevice(filepath, &device))
		return 0;

	if(datasize == 0)
		return 0;

	// stop checking if devices were removed/inserted
	// since we're saving a file
	HaltDeviceThread();

	// halt parsing
	HaltParseThread();

	ShowAction("Saving...");

	while(!written && retry == 1)
	{
		if(!ChangeInterface(device, silent))
			break;

		file = fopen (filepath, "wb");

		if(!file)
		{
			if(silent)
				break;

			retry = ErrorPromptRetry("Error creating file!");
			continue;
		}

		while(written < datasize)
		{
			if(datasize - written > 4096) nextwrite=4096;
			else nextwrite = datasize-written;
			writesize = fwrite (buffer+written, 1, nextwrite, file);
			if(writesize != nextwrite) break; // write failure
			written += writesize;
		}
		fclose (file);

		if(written != datasize) written = 0;

		if(!written)
		{
			unmountRequired[device] = true;
			if(silent) break;
			retry = ErrorPromptRetry("Error saving file!");
		}
	}

	// go back to checking if devices were inserted/removed
	ResumeDeviceThread();

	CancelAction();
	return written;
}

size_t SaveFile(char * filepath, size_t datasize, bool silent)
{
	return SaveFile((char *)savebuffer, filepath, datasize, silent);
}

#include <errno.h>
#include <ogc/es.h>
#include <ogc/lwp_watchdog.h>
#include <ogcsys.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <sys/dir.h>
#include <sys/iosupport.h>
#include <stdarg.h>


void gprintf(const char *str, ...)
{
	char astr[4096];

	va_list ap;
	va_start(ap,str);

	vsprintf( astr, str, ap );

	va_end(ap);

	usb_flush(1);
	//usb_sendbuffer_safe( 1, astr, strlen(astr) );
	usb_sendbuffer( 1, astr, strlen(astr) );
}

char ascii(char s)
{
    if(s < 0x20)
	return '.';
    if(s > 0x7E)
	return '.';
    return s;
}

void hexdump(void *d, int len)
{
    u8 *data;
    int i, off;
    data = (u8*)d;

    gprintf("\n       0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  0123456789ABCDEF");
    gprintf("\n====  ===============================================  ================\n");

    for (off=0; off<len; off += 16)
    {
	gprintf("%04x  ",off);
	for(i=0; i<16; i++)
	    if((i+off)>=len)
		gprintf("   ");
	else
	    gprintf("%02x ",data[off+i]);

	gprintf(" ");
	for(i=0; i<16; i++)
	    if((i+off)>=len)
		gprintf(" ");
	    else
		gprintf("%c",ascii(data[off+i]));
	gprintf("\n");
    }
}
#define DEVICE_NAME "dvd"
#define FST_MAXPATHLEN 255
#define SECTOR_SIZE 0x800

typedef struct {
    u32 dol_offset;
    u32 fst_offset;
    u32 fst_size;
    u32 fst_size2;
} __attribute__ ((packed)) FST_INFO;

typedef struct {
    u8 filetype;
    char name_offset[3];
    u32 fileoffset;
    u32 filelen;
} __attribute__((packed)) FST_ENTRY;

typedef struct
{
    bool inUse;
    u32 offset;
    s32 entry;
} OpenFstItem;

//these are the only memory consumed by mounting this devotab
OpenFstItem openDir;
OpenFstItem openFile;
static u8 *fst_buffer = NULL;

//pointer to the name list at the end of fst_buffer
static char *name_table = NULL;
static FST_ENTRY *fst = NULL;

static u64 last_access = 0;
static s32 dotab_device = -1;

const char *FstName( const FST_ENTRY *entry )
{
    if( entry == &fst[ 0 ] )
	return NULL;

    return (char*)name_table + ( *((u32 *)entry) & 0x00ffffff );
}

static int strcmp_slash( const char *s1, const char *s2 )
{
    while( *s1 == *s2++ )
	if( *s1++ == 0 || *s1 == '/' )
	    return( 0 );

    return( *(const unsigned char*)s1 - *(const unsigned char *)( s2 - 1 ) );
}

static int strlen_slash( const char *s )
{
    int ret = 0;
    while( *s && *s++ != '/' )
	ret++;

    return ret;
}

static u32 NextEntryInFolder( u32 current, u32 directory )
{
    u32 next = ( fst[ current ].filetype ? fst[ current ].filelen : current + 1 );
    if( next < fst[ directory ].filelen )
	return next;

    return 0;
}

static s32 EntryFromPath( const char *path, int d )
{

    if( !strncmp( path, "dvd:/", strlen( "dvd:/" ) ) )
	return EntryFromPath( path + 5, 0 );

    if( !fst[ d ].filetype )
    {
	gprintf("ERROR!!  %s is not a directory\n", FstName( &fst[ d ] ) );
	return -1;
    }

    u32 next = d + 1;

    FST_ENTRY *entry = &fst[ next ];

    while( next )
    {
	//gprintf( "\tcomparing: [ %u ] %s\n", next, FstName( entry ) );
	//does this entry match.
	//strlen_slash is used because if looking for "dvd:/gameboy/" it would return a false positive if it hit "dvd:/gameboy advance/" first
	if( !strcmp_slash( path, FstName( entry ) ) && ( strlen( FstName( entry ) ) == strlen_slash( path ) ) )
	{
	    char *slash = strchr( path, '/' );
	    if( slash && strlen( slash + 1 ) )
	    {
		//we are looking for a file somewhere in this folder
		return EntryFromPath( slash + 1, next );
	    }
	    //this is the actual entry we were looking for
	    return next;
	}

	//find the next entry in this folder
	next = NextEntryInFolder( next, d );
	entry = &fst[ next ];
    }

    //no entry with the given path was found
    return -1;
}

static int _FST_open_r( struct _reent *r, void *fileStruct, const char *path, int flags, int mode )
{
    //gprintf("_FST_open_r( \"%s\"\n", path );
    if( openFile.inUse )
    {
	gprintf("tried to open more than 1 file at a time\n");
	r->_errno = ENOENT;
	return -1;
    }
    openFile.entry = EntryFromPath( path, 0 );
    if( openFile.entry < 0 )
    {
	//gprintf("invalid filename: %s\n", path );
	r->_errno = ENOENT;
	return -1;
    }
    //gprintf("%s\n", FstName( &fst[ openFile.entry ] ) );
    if( fst[ openFile.entry ].filetype )
    {
	gprintf("tried to open a folder: %s\n", path );
	r->_errno = EISDIR;
	return -1;
    }
    openFile.offset = 0;
    openFile.inUse = true;

    return 1;
}

static int _FST_close_r( struct _reent *r, int fd )
{
    //gprintf("close\n");
    if( !openFile.inUse )
    {
	gprintf("tried to close a file that isnt open\n" );
	r->_errno = EBADF;
	return -1;
    }
    memset( &openFile, 0, sizeof( OpenFstItem ) );
    return 0;
}

static int _FST_read_r( struct _reent *r, int fd, char *ptr, size_t len )
{
    //gprintf("read( %p, %u)  current pos: %d\n", ptr, len, openFile.offset );
    if( !openFile.inUse )
    {
	gprintf("read: !openFile.inUse\n");
	r->_errno = EBADF;
	return -1;
    }
    if( openFile.offset >= fst[ openFile.entry ].filelen )
    {
	//gprintf("read: 2\n");
	r->_errno = EOVERFLOW;
	return 0;
    }
    if( len + openFile.offset > fst[ openFile.entry ].filelen )
    {
	//gprintf("read: 3\n");
	r->_errno = EOVERFLOW;
	len = fst[ openFile.entry ].filelen - openFile.offset;
    }
    if( len <= 0 )
    {
	gprintf("read: 4\n");
	return 0;
    }

    if( WDVD_Read_Safe( (u8*)ptr, len, (u64)( ( (u64) fst[ openFile.entry ].fileoffset << 2LL ) + openFile.offset ) ) < 0 )
    {
	gprintf("read: failed\n");
	r->_errno = EIO;
	return -1;
    }

    openFile.offset += len;
    //hexdump( ptr, len + 4 );
    //gprintf("read: ok\tnew pos: %d\n", openFile.offset );

    return len;

}

static off_t _FST_seek_r( struct _reent *r, int fd, off_t pos, int dir )
{
    //gprintf("_FST_seek_r()\n" );

    if( !openFile.inUse )
    {
	gprintf("seek: !openFile.inUse\n");
	r->_errno = EBADF;
	return -1;
    }

    int position;

    switch( dir ) {
	case SEEK_SET:
	    position = pos;
	    break;
	case SEEK_CUR:
	    position = openFile.offset + pos;
	    break;
	case SEEK_END:
	    position = fst[ openFile.entry ].filelen + pos;
	    break;
	default:
	    r->_errno = EINVAL;
	    return -1;
    }

    if( pos > 0 && position < 0 )
    {
	gprintf("seek: pos > 0 && position < 0\n");
	r->_errno = EOVERFLOW;
	return -1;
    }

    if( position < 0 || position > fst[ openFile.entry ].filelen )
    {
	gprintf("seek: shit\n");
	r->_errno = EINVAL;
	return -1;
    }
    openFile.offset = position;

    return position;
}

static void stat_entry_( u32 entry, struct stat *st) {
    st->st_dev = 0x4657;
    st->st_ino = 0;
    st->st_mode = ( fst[ entry ].filetype ? S_IFDIR : S_IFREG) | (S_IRUSR | S_IRGRP | S_IROTH);
    st->st_nlink = 1;
    st->st_uid = 1;
    st->st_gid = 2;
    st->st_rdev = st->st_dev;
    st->st_size = fst[ entry ].filelen;
    st->st_atime = 0;
    st->st_spare1 = 0;
    st->st_mtime = 0;
    st->st_spare2 = 0;
    st->st_ctime = 0;
    st->st_spare3 = 0;
    st->st_blksize = SECTOR_SIZE;
    st->st_blocks = ( fst[ entry ].filelen + SECTOR_SIZE - 1) / SECTOR_SIZE;
    st->st_spare4[0] = 0;
    st->st_spare4[1] = 0;
}

static int _FST_fstat_r(struct _reent *r, int fd, struct stat *st) {
    gprintf("fstat()\n" );

    if( !openFile.inUse )
    {
	r->_errno = EBADF;
	return -1;
    }
    stat_entry_( openFile.entry, st );

    return 0;
}

static int _FST_stat_r(struct _reent *r, const char *path, struct stat *st) {
    //gprintf("stat( \"%s\" )\n", path );
    s32 er = EntryFromPath( path, 0 );
    if( er < 0 )
    {
	r->_errno = ENOENT;
	return -1;
    }
    stat_entry_( er, st );

    return 0;
}

static int _FST_chdir_r(struct _reent *r, const char *path) {
    //gprintf("chdir( \"%s\" )\n", path );
    s32 er = EntryFromPath( path, 0 );
    if( er < 0 )
    {
	gprintf("chdir() bad directory name   %s\n", path );
	r->_errno = ENOTDIR;
	return -1;
    }
    if( !fst[ er ].filetype )
    {
	gprintf("chdir() !filetype\n");
	r->_errno = ENOTDIR;
	return -1;
    }

    memset( &openDir.inUse, 0, sizeof( OpenFstItem ) );
    openDir.entry = er;
    openDir.offset = openDir.entry ? openDir.entry : openDir.entry + 1;
    openDir.inUse = true;

    return 0;
}

static DIR_ITER *_FST_diropen_r(struct _reent *r, DIR_ITER *dirState, const char *path)
{
    //gprintf("dir open( \"%s\" )\n", path );
    if( openDir.inUse )
    {
	gprintf("tried to open more than 1 directory\n");
	r->_errno = ENOENT;
	return NULL;
    }
    if( !strcmp( path, "dvd:/" ) )
	openDir.entry = 0;
    else
	openDir.entry = EntryFromPath( path, 0 );

    if( openDir.entry < 0 )
    {
	gprintf("bad filename\n");
	r->_errno = ENOENT;
	return NULL;
    }

    if( !fst[ openDir.entry ].filetype )
    {
	gprintf("not a directory: %s\n", path );
	r->_errno = ENOTDIR;
	return NULL;
    }
    //gprintf("returns %d: %s\n", openDir.entry, FstName( &fst[ openDir.entry ] ) );


    openDir.offset = openDir.entry ? openDir.entry : openDir.entry + 1;

    openDir.inUse = true;
    dirState = (DIR_ITER*)&openDir;

    return dirState;
}

static int _FST_dirreset_r( struct _reent *r, DIR_ITER *dirState )
{
    //gprintf("dirreset()\n" );
    if( !openDir.inUse )
    {
	r->_errno = EBADF;
	return -1;
    }
    openDir.offset = openDir.entry ? openDir.entry : openDir.entry + 1;
    return 0;
}

static int _FST_dirnext_r( struct _reent *r, DIR_ITER *dirState, char *filename, struct stat *st )
{
    if( !openDir.inUse )
    {
	gprintf("dirnext(): !inUse\n");
	r->_errno = EBADF;
	return -1;
    }
    if( openDir.entry && openDir.offset == openDir.entry )//this is the first entry in a directory, and not the root directory
    {
	strcpy( filename, ".." );
	stat_entry_( fst[ openDir.offset ].fileoffset, st );
	openDir.offset = fst[ openDir.entry ].filelen > openDir.entry + 1 ? openDir.offset + 1 : 0;
	return 0;
    }
    if( !openDir.offset )//no more entries in this directory
    {
	r->_errno = ENOENT;
	return -1;
    }

    //gprintf("dirnext: %d: %s\n", openDir.offset, FstName( &fst[ openDir.offset ] ) );

    strncpy( filename, FstName( &fst [ openDir.offset ] ), FST_MAXPATHLEN - 1 );
    stat_entry_( openDir.offset, st );
    openDir.offset = NextEntryInFolder( openDir.offset, openDir.entry );

    return 0;
}

static int _FST_dirclose_r( struct _reent *r, DIR_ITER *dirState )
{
    if( !openDir.inUse )
    {
	gprintf("dirclose(): !openDir.inUse\n");
	r->_errno = EBADF;
	return -1;
    }

    memset( &openDir.inUse, 0, sizeof( OpenFstItem ) );

    return 0;
}

static int _FST_statvfs_r(struct _reent *r, const char *name, struct statvfs *buf)
{
    if(!buf)
	return -1;

    memset(buf, 0,  sizeof(struct statvfs));
    return 0;
}

static const devoptab_t dotab_fst = {
    DEVICE_NAME,
    sizeof(OpenFstItem),
    _FST_open_r,
    _FST_close_r,
    NULL,
    _FST_read_r,
    _FST_seek_r,
    _FST_fstat_r,
    _FST_stat_r,
    NULL,
    NULL,
    _FST_chdir_r,
    NULL,
    NULL,
    sizeof(OpenFstItem),
    _FST_diropen_r,
    _FST_dirreset_r,
    _FST_dirnext_r,
    _FST_dirclose_r,
    _FST_statvfs_r
};

static bool read_disc() {
    //gprintf("read_disc()\n");

    s32 ret;
    FST_INFO fst_info;
    u32 name_table_offset;

    //find FST inside partition
    ret = WDVD_Read( (u8*)&fst_info, sizeof( FST_INFO ), 0x420LL );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_info ): %d\n", ret );
	goto error;
    }
    fst_info.fst_offset <<= 2;
    fst_info.fst_size <<= 2;
    //gprintf("fst offset: %08x\tsize: %08x\n", fst_info.fst_offset, fst_info.fst_size );
    //hexdump( &fst_info, 0x20 );

    fst_buffer = (u8*)memalign( 32, fst_info.fst_size );
    if( !fst_buffer ) goto error;

    //read fst into memory
    ret = WDVD_Read( fst_buffer, fst_info.fst_size, fst_info.fst_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_buffer ): %d\n", ret );
	goto error;
    }

    //set the pointers
    fst = (FST_ENTRY *)fst_buffer;
    name_table_offset = fst->filelen * 0xC;
    name_table = (char *)( fst_buffer + name_table_offset );

    //clear the openFile and openDir entries
    memset( &openDir, 0, sizeof( OpenFstItem ) );
    memset( &openFile, 0, sizeof( OpenFstItem ) );

    //gprintf("read_disc ok\n");
    return true;

    error:
    gprintf("read_disc fail\n");
    if( fst_buffer )
	free( fst_buffer );

    fst = NULL;
    name_table = NULL;

    return false;
}

static bool alreadyMounted = false;

bool FSTIsMounted()
{
    return alreadyMounted;
}

bool FST_Mount()
{
    //gprintf("FST_Mount()\n");

    if( alreadyMounted )
	FST_Unmount();

    bool success = read_disc() && ( dotab_device = AddDevice( &dotab_fst ) ) >= 0;
    if( success )
	last_access = gettime();
    else
	FST_Unmount();

    alreadyMounted = true;
    //gprintf("okey dokey\n");
    return success;
}

bool FST_Unmount()
{
    //gprintf("FST_Unmount()\n");

    alreadyMounted = false;
    if( fst_buffer )
    {
	free( fst_buffer );
	fst_buffer = NULL;
    }
    name_table = NULL;
    fst = NULL;

    last_access = 0;
    if (dotab_device >= 0) {
	dotab_device = -1;
	return !RemoveDevice(DEVICE_NAME ":");
    }

    return true;
}

u64 FST_LastAccess() {
    return last_access;
}

#define IOCTL_DI_READID		0x70
#define IOCTL_DI_READ		0x71
#define IOCTL_DI_WAITCVRCLOSE	0x79
#define IOCTL_DI_GETCOVER	0x88
#define IOCTL_DI_RESET		0x8A
#define IOCTL_DI_OPENPART	0x8B
#define IOCTL_DI_CLOSEPART	0x8C
#define IOCTL_DI_UNENCREAD	0x8D
#define PTABLE_OFFSET	0x40000

#define READ_BUFFER_ZISE

static u32 inbuf[ 8 ]  ATTRIBUTE_ALIGN( 32 );
static u32 outbuf[ 8 ] ATTRIBUTE_ALIGN( 32 );

#define CACHE_END ( cacheStart + CACHE_SIZE )
#define CACHE_SIZE 0x800
static u8 readCache[ CACHE_SIZE ] ATTRIBUTE_ALIGN( 32 );
static u64 cacheStart = 0;


static const char di_fs[] ATTRIBUTE_ALIGN( 32 ) = "/dev/di";
static s32 di_fd = -1;

s32 WDVD_Init(void) {
    if (di_fd < 0) {
	di_fd = IOS_Open(di_fs, 0);
	if (di_fd < 0)
	    return di_fd;
    }

    return 0;
}

s32 WDVD_Close(void) {
    if (di_fd >= 0) {
	IOS_Close(di_fd);
	di_fd = -1;
    }

    return 0;
}

s32 WDVD_Reset(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_RESET << 24;
    inbuf[1] = 1;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_RESET, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ReadDiskId(void *id) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_READID << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_READID, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    if (ret == 1) {
	memcpy(id, outbuf, sizeof(dvddiskid));
	return 0;
    }

    return -ret;
}

s32 WDVD_OpenPartition(u64 offset) {
    u8 *vector = NULL;

    u32 *buffer = NULL;
    s32 ret;

    buffer = (u32 *)memalign(32, 0x5000);
    if (!buffer)
	return -1;

    vector = (u8 *)buffer;

    memset(buffer, 0, 0x5000);

    buffer[0] = (u32)(buffer + 0x10);
    buffer[1] = 0x20;
    buffer[3] = 0x024A;
    buffer[6] = (u32)(buffer + 0x380);
    buffer[7] = 0x49E4;
    buffer[8] = (u32)(buffer + 0x360);
    buffer[9] = 0x20;

    buffer[(0x40 >> 2)]     = IOCTL_DI_OPENPART << 24;
    buffer[(0x40 >> 2) + 1] = offset >> 2;

    ret = IOS_Ioctlv(di_fd, IOCTL_DI_OPENPART, 3, 2, (ioctlv *)vector);

    free(buffer);

    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ClosePartition(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_CLOSEPART << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_CLOSEPART, inbuf, sizeof(inbuf), NULL, 0);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_UnencryptedRead(void *buf, u32 len, u64 offset) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_UNENCREAD << 24;
    inbuf[1] = len;
    inbuf[2] = (u32)(offset >> 2);

    ret = IOS_Ioctl(di_fd, IOCTL_DI_UNENCREAD, inbuf, sizeof(inbuf), buf, len);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_Read(void *buf, u32 len, u64 offset) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_READ << 24;
    inbuf[1] = len;
    inbuf[2] = (u32)(offset >> 2);

    ret = IOS_Ioctl(di_fd, IOCTL_DI_READ, inbuf, sizeof(inbuf), buf, len);
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_ReadIntoCache( u64 offset )
{
    u8 off = 0;
    while( ( ( ( offset - off ) >> 2ull ) << 2ull ) != offset - off )
    {
	off++;
    }

    s32 ret = WDVD_Read( &readCache, CACHE_SIZE, offset - off );
    if( ret < 0 )
    {
	gprintf("WDVD_Read(): %d\n", ret);
	return ret;
    }
    cacheStart = offset - off;

    return 0;
}

s32 WDVD_Read_Safe( void *buf, u32 len, u64 offset )
{
    //gprintf("WDVD_Read_Safe( %p, %08x, %016llx )\n", buf, len, offset );
    u8 off = 0;
    while( ( ( ( offset - off ) >> 2ull ) << 2ull ) != offset - off )
    {
	off++;
    }
    if( !off && len >= CACHE_SIZE ) //nice happy aligned offset and not just 4 bytes at a time, just read it and be done
	return WDVD_Read( buf, len, offset );

    //for the "read 1 byte at a time" shit
    u32 haveRead = 0;
    u32 howFarIntoTheCacheIsIt = 0;
    u32 howMuchOfTheCachDoINeedToCopy = 0;
    while( haveRead < len )
    {
	if( offset >= cacheStart && offset < cacheStart + CACHE_SIZE )
	{
	    howFarIntoTheCacheIsIt = (u32)( (u64)offset - (u64)cacheStart );
	    howMuchOfTheCachDoINeedToCopy = MIN( CACHE_END - howFarIntoTheCacheIsIt, len );

	    memcpy( buf + haveRead, (u8*)(&readCache) + howFarIntoTheCacheIsIt, howMuchOfTheCachDoINeedToCopy );
	}
	else
	{
	    s32 ret = WDVD_ReadIntoCache( offset + haveRead );
	    if( ret < 0 )
		return ret;
	}
	haveRead += howMuchOfTheCachDoINeedToCopy;
    }
    return 0;
}

s32 WDVD_WaitForDisc(void) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_WAITCVRCLOSE << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_WAITCVRCLOSE, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    return (ret == 1) ? 0 : -ret;
}

s32 WDVD_GetCoverStatus(u32 *status) {
    s32 ret;

    memset(inbuf, 0, sizeof(inbuf));

    inbuf[0] = IOCTL_DI_GETCOVER << 24;

    ret = IOS_Ioctl(di_fd, IOCTL_DI_GETCOVER, inbuf, sizeof(inbuf), outbuf, sizeof(outbuf));
    if (ret < 0)
	return ret;

    if (ret == 1) {
	memcpy(status, outbuf, sizeof(u32));

	return 0;
    }

    return -ret;
}

s32 WDVD_FindPartition(u64 *outbuf)
{
    u64 offset = 0, table_offset = 0;

    u32 cnt, nb_partitions;
    s32 ret;
    u32 *buffer = NULL;

    buffer = (u32 *)memalign(32, 0x20);
    if (!buffer)
	return -1;

    ret = WDVD_UnencryptedRead(buffer, 0x20, PTABLE_OFFSET);
    if (ret < 0)
    {
	free( buffer );
	return ret;
    }

    nb_partitions = buffer[0];
    table_offset  = buffer[1] << 2;

    ret = WDVD_UnencryptedRead(buffer, 0x20, table_offset);
    if (ret < 0)
    {
	free( buffer );
	return ret;
    }

    for (cnt = 0; cnt < nb_partitions; cnt++) {
	u32 type = buffer[cnt * 2 + 1];

	if (!type)
	{
	    offset = buffer[cnt * 2] << 2;
	    break;
	}
    }

    free( buffer );

    if (!offset)
	return -1;

    *outbuf = offset;

    return 0;
}

s32 WDVD_OpenDataPartition()
{
    u64 dataPartitionOffset;
    u32 status = 0;

    s32 ret = WDVD_GetCoverStatus( &status );
    if( ret < 0 )
    {
	gprintf("WDVD_GetCoverStatus(): %d\n", ret );
	return ret;
    }
    if( status != 2 )
    {
	gprintf("discstatus: %d\n", status );
	return -1;
    }

    ret = WDVD_Reset();
    if( ret < 0 )
    {
	gprintf("WDVD_Reset(): %d\n", ret );
	return ret;
    }

    ret = WDVD_ReadDiskId( (void*)0x80000000 );
    if( ret < 0 )
    {
	gprintf("WDVD_ReadDiskId(): %d\n", ret );
	return ret;
    }

    ret = WDVD_FindPartition( &dataPartitionOffset );
    if( ret < 0 )
    {
	gprintf("WDVD_FindPartition(): %d\n", ret );
	return ret;
    }

    ret = WDVD_OpenPartition( dataPartitionOffset );
    if( ret < 0 )
    {
	gprintf("WDVD_OpenPartition( %016llx ): %d\n", dataPartitionOffset, ret );
	return ret;
    }

    //gprintf("opened partition @ %016llx\n", dataPartitionOffset );

    return 0;
}

s32 WDVD_GetMainDol( void *outbuf, u32* size )
{
    s32 ret;
    FST_INFO fst_info;
    //get dol offset
    ret = WDVD_Read( (u8*)&fst_info, sizeof( FST_INFO ), 0x420LL );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( fst_info ): %d\n", ret );
	return -1;
    }
    fst_info.dol_offset <<= 2;

    //gprintf("dol offset: %08x\n", fst_info.dol_offset );

    //determine dol size
    ret = WDVD_Read( (u8*)outbuf, 0x100, fst_info.dol_offset );
    if( ret < 0 )
    {
	gprintf("WDVD_Read( dol header ): %d\n", ret );
	return -2;
    }
    u32 max = 0;
    u32 i;
    for (i = 0; i < 18; i++) {
	u32 offset = *(u32 *)( outbuf + (i * 4));
	u32 size = *(u32 *)( outbuf + (i * 4) + 0x90);
	if ((offset + size) > max) max = offset + size;
    }
    //gprintf("dol size: %08x\n", max );

    //read the dol
    ret = WDVD_Read( (u8*)outbuf, max, fst_info.dol_offset );
    if( ret < 0 )
    {
	printf("WDVD_Read( main.dol ): %d\n", ret );
	return -3;
    }

    *size = max;
    //gprintf("dol read ok\n", max );

    return 0;
}

